# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import argparse
import codecs
import os
import signal
import sys
import threading
import time

# Save PEP 3122!
if "." in __name__:
    from . import common
else:
    import common

sys.stderr = codecs.getwriter("utf-8")(sys.stderr.detach())
sys.stdout = codecs.getwriter("utf-8")(sys.stdout.detach())

E_PYTHON_VERSION = """
Your version of the python programming language is too old to run saxo. Use a
newer version if available. Otherwise you can get a newer version from here:

    http://www.python.org/

Or install one using your system's package manager. If you are using OS X for
example, python3 can be installed using homebrew:

    http://mxcl.github.io/homebrew/
"""

try: import sqlite3
except ImportError:
    print("Error: sqlite3 is not installed", file=sys.stderr)
    print("Please build Python against the sqlite libraries", file=sys.stderr)
    sys.exit(1)
else:
    if not sqlite3.threadsafety:
        print("Error: Your sqlite3 is not thread-safe", file=sys.stderr)
        sys.exit(1)

if sys.version_info < (3, 3):
    common.error("requires python 3.3 or later", E_PYTHON_VERSION)

def action(function):
    action.names[function.__name__] = function
    return function
action.names = {}

lock = threading.Lock()

def debug(*args, **kargs):
    with lock:
        try:
            print(*args, **kargs)
            sys.stdout.flush()
        except BrokenPipeError:
            sys.exit()

# Options

def version(args, v):
    print("saxo %s" % v)

USAGE = """
Usage:
    saxo -v
    saxo create [ directory ]
    saxo [ -f ] [ --log ] start [ directory ]
    saxo stop [ directory ]
    saxo active [ directory ]

Try `saxo -h` for more detailed usage
"""

def usage(args, v):
    version(args, v)
    print(USAGE.rstrip())

HELP = """
Usage:

    saxo -v / --version
        Show the current saxo version

    saxo create [ directory ]
        Create a default configuration file

    saxo [ Options ] start [ directory ]
        Starts a bot. Options:

            -f / --foreground
                Don't run the bot as a daemon

            -l / --log
                Save output to a log file in saxo base

    saxo stop [ directory ]
        Stops a bot

    saxo active [ directory ]
        Shows whether a bot is active
"""

def help(args, v):
    version(args, v)
    print(HELP.rstrip())

def base_option(args):
    if args.directory is None:
        return os.path.expanduser("~/.saxo")
    return args.directory

# Actions

# eval $(./saxo bash)
@action
def bash(arg):
    # Save PEP 3122!
    if "." in __name__:
        from . import saxo
    else:
        import saxo

    import pipes

    path = pipes.quote(os.path.join(saxo.path, "commands"))
    print('export PATH=$PATH:%s' % path)

    pythonpath = pipes.quote(saxo.path)
    if "PYTHONPATH" in os.environ:
        print('export PYTHONPATH=$PYTHONPATH:%s' % pythonpath)
    else:
        print('export PYTHONPATH=%s' % pythonpath)

@action
def create(args):
    # Save PEP 3122!
    if "." in __name__:
        from . import create
    else:
        import create

    # def default(base=None)
    # — create.py
    create.default(args.directory)
    return 0

@action
def shell(args):
    base = base_option(args)

    # Save PEP 3122!
    if "." in __name__:
        from .saxo import path as saxo_path
    else:
        from saxo import path as saxo_path

    def subshell(base, commands):
        path = os.environ.get("PATH", "")
        os.environ["PATH"] = commands + os.pathsep + path
        os.environ["PYTHONPATH"] = saxo_path
        os.environ["SAXO_SHELL"] = "1"
        os.environ["SAXO_BASE"] = base
        os.environ["SAXO_BOT"] = "saxo"
        os.environ["SAXO_COMMANDS"] = commands
        os.environ["SAXO_NICK"] = os.environ["USER"]
        os.environ["SAXO_SENDER"] = os.environ["USER"]
        shell = os.environ.get("SHELL", "sh")
        os.system(shell)

    commands = os.path.join(base, "commands")
    subshell(base, commands)
    return 0

@action
def console(args):
    # TODO: Reduce code duplication
    base = base_option(args)

    # Save PEP 3122!
    if "." in __name__:
        from . import saxo
    else:
        import saxo

    path = os.environ.get("PATH", "")
    commands = os.path.join(base, "commands")

    # TODO: Reload symlinks
    # e.g. the pip-installation test changes them

    os.environ["PATH"] = commands + os.pathsep + path
    os.environ["PYTHONPATH"] = saxo.path
    os.environ["SAXO_SHELL"] = "1"
    os.environ["SAXO_BASE"] = base
    os.environ["SAXO_BOT"] = "saxo"
    os.environ["SAXO_COMMANDS"] = commands
    os.environ["SAXO_NICK"] = os.environ["USER"]
    os.environ["SAXO_SENDER"] = os.environ["USER"]

    def do(cmd, arg):
        import subprocess
        cmd = os.path.join(os.environ["SAXO_COMMANDS"], cmd)
        try: octets = subprocess.check_output([cmd, arg])
        except Exception as err:
            return "Error: %s" % err
        return octets.decode("utf-8")

    while True:
        sys.stdout.write("> ")
        sys.stdout.flush()
        try: i = sys.stdin.readline()
        except: break

        i = i.rstrip("\r\n")
        if " " in i:
            cmd, arg = i.split(" ", 1)
        else:
            cmd, arg = i, ""

        if cmd.startswith("."):
            saxo.client(cmd[1:], *eval("(%s)" % arg))

        elif cmd.startswith("!"):
            print(do(cmd[1:], arg).rstrip("\r\n"))

        elif cmd == "saxo.path":
            print(saxo.path)

        elif cmd == "saxo.base":
            print(base)

        elif not cmd:
            break

        else:
            print("Unknown command:", repr(cmd))

@action
def log(args):
    base = base_option(args)
    print(os.path.join(base, "log"))

@action
def start(args):
    import json

    # Quit any previously connected instances
    if "." in __name__:
        from .saxo import client, data
    else:
        from saxo import client, data

    base = base_option(args)
    try: client("quit", base=base)
    except FileNotFoundError as err:
        ...
    except ConnectionRefusedError as err:
        ...
    except PermissionError as err:
        debug("Error: Unable to connect to internal socket", file=sys.stderr)
        debug("Check permissions on the config dir", file=sys.stderr)
        sys.stderr.flush()
        sys.exit(1)
    else:
        debug("Warning: Client may already have been running!")
        sys.stdout.flush()

    if not os.path.isdir(base):
        debug("Error: The saxo config dir does not exist", file=sys.stderr)
        debug("Location: %s" % base, file=sys.stderr)
        sys.stderr.flush()
        sys.exit(1)

    pidfile = os.path.join(base, "pid")
    if not args.foreground:
        # Save PEP 3122!
        if "." in __name__:
            from . import daemon
        else:
            import daemon

        if args.log:
            log = os.path.join(base, "log")
            if os.path.exists(log):
                import shutil
                modified = time.gmtime(os.path.getmtime("saxo"))
                log2 = time.strftime("log-%Y%m%d-%H%M%S.txt", modified)
                shutil.move(log, os.path.join(base, log2))
            output = open(log, "w")
        else:
            output = open(os.devnull, "w")

        daemon.start(pidfile, output)
    else:
        # This is duplicated variatim from daemon.py
        import atexit
        with open(pidfile, "w") as f:
            f.write(str(os.getpid()) + "\n")

        def delete_pidfile():
            if os.path.isfile(pidfile):
                os.remove(pidfile)
        atexit.register(delete_pidfile)

    if args.action == "start":
        # Otherwise you get recursion if args.action == "restart"
        os.environ["SAXO_BASE"] = base
        sys_args = json.dumps(args.original)
        data("args", sys_args, command="script.py", check=False)
    
    # Save PEP 3122!
    if "." in __name__:
        from . import irc
    else:
        import irc

    irc.start(base)
    return 0

@action
def restart(args):
    import json

    base = base_option(args)
    os.environ["SAXO_BASE"] = base

    if "." in __name__:
        from .saxo import data, version
    else:
        from saxo import data, version

    sys_args = json.loads(data("args", command="script.py", check=False))
    if sys_args[0] == "restart":
        debug("Recursion detected!", sys_args)
        debug("Not restarting")
        sys.exit(1)

    debug("Stopping existing instance...")
    stop(args)
    argv = [sys.argv[0]] + sys_args
    debug("Restarting using:", argv)
    main(argv, version)

@action
def status(args):
    base = base_option(args)

    # Quit any previously connected instances
    if "." in __name__:
        from .saxo import client
    else:
        from saxo import client

    try: client("noop", base=base)
    except FileNotFoundError as err:
        debug("not running")
    except ConnectionRefusedError as err:
        debug("not running")
    else:
        debug("running")

@action
def stop(args):
    base = base_option(args)

    if "." in __name__:
        from .saxo import client
    else:
        from saxo import client

    try: client("quit", base=base)
    except: ...

    pidfile = os.path.join(base, "pid")
    if not os.path.exists(pidfile):
        common.error("There is no bot currently running")

    with open(pidfile, encoding="ascii") as f:
        text = f.read()
        pid = int(text.rstrip("\n"))

    # TODO: Make this less crude
    # This can give a ProcessLookupError: [Errno 3] No such process
    try: os.kill(pid, signal.SIGTERM)
    except ProcessLookupError:
        ...
    # os.kill(pid, signal.SIGKILL)

    return 0

@action
def test(args):
    if args.directory is not None:
        common.error("Tests cannot be run in conjunction with a directory")

    import queue
    import shutil
    import socket
    import subprocess
    import tempfile

    # Save PEP 3122!
    if "." in __name__:
        from . import saxo
    else:
        import saxo

    saxo_script = sys.modules["__main__"].__file__
    saxo_test_server = os.path.join(saxo.path, "test", "server.py")

    tmp = tempfile.mkdtemp()
    outgoing = queue.Queue()

    if not sys.executable:
        common.error("Couldn't find the python executable")

    if not os.path.isdir(tmp):
        common.error("There is no %s directory" % tmp)

    print("python executable:", sys.executable)
    print("saxo path:", saxo.path)
    print("saxo script:", saxo_script)
    print("saxo test server:", saxo_test_server)
    print()

    def run_server():
        server = subprocess.Popen([sys.executable, "-u", saxo_test_server],
            stdout=subprocess.PIPE)

        for line in server.stdout:
            line = line.decode("utf-8", "replace")
            line = line.rstrip("\n")
            outgoing.put("S: " + line)

        outgoing.put("Server finished")

    def run_client():
        saxo_test = os.path.join(tmp, "saxo-test")
        outgoing.put("Running in %s" % saxo_test)

        cmd = [sys.executable, saxo_script, "create", saxo_test]
        code = subprocess.call(cmd)
        if code:
            print("Error creating the client configuration")
            sys.exit(1)

        test_config = os.path.join(saxo.path, "test", "config")
        saxo_test_config = os.path.join(saxo_test, "config")
        with open(test_config) as f:
            with open(saxo_test_config, "w") as w:
                for line in f:
                    line = line.replace("localhost", socket.gethostname())
                    w.write(line)
        # shutil.copy2(test_config, saxo_test_config)

        client = subprocess.Popen([sys.executable, "-u",
                saxo_script, "-f", "start", saxo_test],
            stdout=subprocess.PIPE)

        for line in client.stdout:
            line = line.decode("utf-8", "replace")
            line = line.rstrip("\n")
            outgoing.put("C: " + line)

        manifest01 = {"commands", "config", "database.sqlite3",
            "pid", "plugins"}
        manifest02 = manifest01 | {"client.sock"}

        if set(os.listdir(saxo_test)) == manifest01:
            shutil.rmtree(saxo_test)
        elif set(os.listdir(saxo_test)) == manifest02:
            outgoing.put("Warning: client.sock had not been removed")
            shutil.rmtree(saxo_test)
        else:
            outgoing.put("Refusing to delete the saxo test directory")
            outgoing.put("Data was found which does not match the manifest")
            outgoing.put(saxo_test)

    common.thread(run_server)
    common.thread(run_client)

    error = False
    completed = False
    client_buffer = []
    while True:
        line = outgoing.get()

        if line.startswith("S: "):
            print(line)
            if line.startswith("S: ERROR"):
               error = True
            if line.startswith("S: Tests complete"):
               completed = True
            if not line.startswith("S: Test"):
                for c in client_buffer:
                    print(c)
            del client_buffer[:]

        elif line.startswith("C: "):
            client_buffer.append(line)

        else:
            print(line)

        sys.stdout.flush()

        if line == "Server finished":
            break

    if not os.listdir(tmp):
        os.rmdir(tmp)
    else:
        print("Warning: Did not remove:", tmp)

    if completed and (not error):
        sys.exit(0)
    else:
        sys.exit(1)

@action
def soft(args):
    import site

    if args.directory == "on":
        # Save PEP 3122!
        if "." in __name__:
            from .saxo import path as saxo_path
        else:
            from saxo import path as saxo_path

        if not os.path.isdir(site.USER_SITE):
            os.makedirs(site.USER_SITE)
            debug("Warning: Created", site.USER_SITE)
        saxo_pth = os.path.join(site.USER_SITE, "saxo.pth")
        if not os.path.exists(saxo_pth):
            with open(saxo_pth, "w") as f:
                f.write(saxo_path)
            debug("Created saxo.pth in", site.USER_SITE)
        else:
            debug("Error: %s already exists" % saxo_pth, file=sys.stderr)
            sys.exit(1)
    elif args.directory == "off":
        saxo_pth = os.path.join(site.USER_SITE, "saxo.pth")
        if os.path.isfile(saxo_pth):
            os.remove(saxo_pth)
        else:
            debug("Error: %s not found" % saxo_pth, file=sys.stderr)
            sys.exit(1)
        debug("Warning: May also need to remove %s" % site.USER_SITE)
    else:
        debug("Error: Was expecting 'on' or 'off'", file=sys.stderr)
        sys.exit(1)

@action
def info(args):
    # Save PEP 3122!
    if "." in __name__:
        from .saxo import path as saxo_path
    else:
        from saxo import path as saxo_path
    directory = args.directory
    args.directory = None
    base = base_option(args)
    args.directory = directory

    def check(path):
        if os.path.isdir(path):
            print(path)
        else:
            debug("-", file=sys.stderr)
            sys.exit(1)

    if args.directory == "base":
        check(base)
    elif args.directory == "path":
        check(saxo_path)
    elif args.directory == "base.commands":
        check(os.path.join(base, "commands"))
    elif args.directory == "path.commands":
        check(os.path.join(saxo_path, "commands"))
    else:
        debug("?", file=sys.stderr)
        sys.exit(1)

def main(argv, v):
    # NOTE: No default for argv, because what script name would we use?
    description = "Control saxo irc bot instances"
    parser = argparse.ArgumentParser(description=description, add_help=False)

    parser.add_argument("-f", "--foreground", action="store_true",
        help="run in the foreground instead of as a daemon")

    parser.add_argument("-h", "--help", action="store_true",
        help="show a short help message")

    parser.add_argument("-o", "--output", metavar="filename",
        help="removed command: use -l / --log with no arg instead")

    parser.add_argument("-l", "--log", action="store_true",
        help="log to the standard log file")

    parser.add_argument("-v", "--version", action="store_true",
        help="show the current saxo version")

    parser.add_argument("action", nargs="?",
        help="use --help to show the available actions")

    parser.add_argument("directory", nargs="?",
        help="the path to the saxo configuration directory")

    args = parser.parse_args(argv[1:])
    args.original = argv[1:]

    if args.output:
        print("Sorry, -o / --output has been removed!")
        print("Use -l / --log instead, with no arguments")
        sys.exit(2)

    if args.help:
        help(args, v)

    elif args.version:
        version(args, v)

    elif args.action:
        if args.action in action.names:
            code = action.names[args.action](args)
            if isinstance(code, int):
                sys.exit(code)
        elif args.action.startswith("."):
            # Save PEP 3122!
            if "." in __name__:
                from . import saxo
            else:
                import saxo

            def do(cmd, arg):
                import subprocess
                cmd = os.path.join(saxo.path, "commands", cmd)
                try: octets = subprocess.check_output([cmd, arg])
                except Exception as err:
                    raise err # return "Error: %s" % err
                return octets.decode("utf-8").rstrip("\r\n")
            print(do(args.action[1:], args.directory or ""))
        else:
            common.error("unrecognised action: %s" % args.action, code=2)

    else:
        usage(args, v)



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import os
import queue
import socket
import sys
import time

# Save PEP 3122!
if "." in __name__:
    from . import sqlite
    from . import common
else:
    import sqlite
    import common

incoming = queue.Queue()

class Scheduler(object):
    def __init__(self, client):
        self.client = client
        self.duration = 0.5
        self.connections = 0
        self.connected = False
        self.running = False

    def message(self, msg):
        self.client.put(("message", "Scheduler: %s" % msg))

    # TODO: Make a monotonic version of time.time()
    def tick(self):
        start = time.time()
        # Check for new scheduled commands
        # TODO: New periodic commands
        while True:
            try: a, b = incoming.get(timeout=self.duration / 6)
            except queue.Empty:
                break
            else:
                if a == "connected":
                    self.connections += 1
                    self.message("connected (%s)" % self.connections)
                    self.connected = True
                elif a == "disconnected":
                    self.message("disconnected, and stopped")
                    self.connected = False
                    self.running = False
                elif a == "start":
                    if self.connected:
                        self.message("started at tick %s" % start)
                        self.running = True
                elif a == "stop":
                    self.message("stopped")
                    self.running = False
                elif a == "schedule.add":
                    if "saxo_schedule" in self.db:
                        self.db["saxo_schedule"].insert(b)
                else:
                    self.message("unknown instruction: %s" % a)

                elapsed = time.time() - start
                if elapsed > (self.duration / 3):
                    break

        if self.connected and self.running:
            # Periodic commands
            if "saxo_periodic" in self.db:
                deletions = []
                additions = []
                periodic = self.db["saxo_periodic"].rows(order="recent")
                for (name, period, recent, command, args) in periodic:
                    # find next %0 point after start
                    unixtime = recent - (recent % period) + period
                    if unixtime > int(start):
                        continue
                    cmd = command.decode("ascii")
                    self.client.put((cmd,) + common.b64unpickle(args))
                    deletions.append((name, period, recent, command, args))
                    additions.append((name, period, int(start), command, args))

                for deletion in deletions:
                    del self.db["saxo_periodic"][deletion]
                for addition in additions:
                    self.db["saxo_periodic"].insert(addition)

            # Scheduled commands
            if "saxo_schedule" in self.db:
                deletions[:] = []
                schedule = self.db["saxo_schedule"].rows(order="unixtime")
                for (unixtime, command, args) in schedule:
                    if unixtime > start:
                        break
                    cmd = command.decode("ascii")
                    self.client.put((cmd,) + common.b64unpickle(args))
                    deletions.append((unixtime, command, args))

                for deletion in deletions:
                    del self.db["saxo_schedule"][deletion]

        elapsed = time.time() - start
        if elapsed < self.duration:
            time.sleep(self.duration - elapsed)

        return True

    def tock(self):
        ...

    def start(self, base):
        database_filename = os.path.join(base, "database.sqlite3")
        with sqlite.Database(database_filename) as self.db:
            self.message("initialised, waiting for instructions")
            while self.tick():
                self.tock()



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import os.path
import saxo
import time

@saxo.setup
def setup(irc):
    path = os.path.join(irc.base, "database.sqlite3")
    with saxo.database(path) as db:
        if "saxo_seen" not in db:
            db["saxo_seen"].create(
                ("nick", "TEXT PRIMARY KEY"),
                ("unixtime", int),
                ("channel", str))
        # TODO: Drop nicknames seen more than a year ago

        if "saxo_private" not in db:
            db["saxo_private"].create(
                ("channel", "TEXT PRIMARY KEY"))

@saxo.event("PRIVMSG")
def record(irc):
    if irc.sender.startswith("#"):
        path = os.path.join(irc.base, "database.sqlite3")
        with saxo.database(path) as db:
            if "saxo_seen" in db:
                # TODO: db["saxo_seen"].replace
                command = "INSERT OR REPLACE INTO saxo_seen" + \
                    " (nick,unixtime,channel) VALUES(?,?,?)"
                db.execute(command, irc.nick, int(time.time()), irc.sender)
                db.commit()



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import atexit
import collections
import configparser
import imp
import importlib
import os.path
import queue
import re
import signal
import socket
import subprocess
import sys
import threading
import time

# Save PEP 3122!
if "." in __name__:
    from . import common
    from . import scheduler
    from . import sqlite
    from .saxo import path as saxo_path
    from .saxo import version as saxo_version
else:
    import common
    import scheduler
    import sqlite
    from saxo import path as saxo_path
    from saxo import version as saxo_version

lock = threading.Lock()

def debug(*args, **kargs):
    with lock:
        try:
            print(*args, **kargs)
            sys.stdout.flush()
        except BrokenPipeError:
            sys.exit()

def exit(code):
    # TODO: Sock removal code
    try: sys.exit(code)
    finally: os._exit(code)
    # TODO: Sometimes sys.exit doesn't work, not sure why

# List of threads:
# client.receive
# client.send
# every plugin function
# every process command, as a communication wrapper
# scheduler
# serve.listen
# every serve.connection instance

incoming = queue.Queue()
outgoing = queue.Queue()

regex_optional_prefix = re.compile(r"(?::([^! ]*)!?([^@ ]*)@?([^ ]*))?")
regex_parameter = re.compile(r"((?:(?<= :)[^\r\n]*)|(?:[^: \r\n][^ \r\n]*))")

def parse(octets):
    text = octets.decode("utf-8", "replace")

    match_prefix = regex_optional_prefix.match(text)
    params = regex_parameter.findall(text[match_prefix.end():])
    return match_prefix.groups(), params[0], params[1:]

class Message(object):
    def __init__(self, saxo, octets):
        prefix, command, parameters = parse(octets)

        self.base = saxo.base[:]
        self.config = saxo.config_cache.copy()

        self.nick = prefix[0]
        self.user = prefix[1]
        self.host = prefix[2]

        if self.nick and self.user and self.host:
            self.prefix = self.nick + "!" + self.user + "@" + self.host

        self.command = command
        self.parameters = parameters
        self.blocked = False

        def send(*args):
            saxo.send(*args)
        self.send = send

        def msg(*args):
            saxo.send("PRIVMSG", *args)
        self.msg = msg

        if self.command == "PRIVMSG":
            self.sender = self.parameters[0]
            self.text = self.parameters[1]
            self.private = self.sender == self.config["nick"]

            if self.private:
                self.sender = self.nick

            if saxo.address:
                # TODO: Why was this 498 for duxlot?
                self.limit = 493 - len(self.sender + saxo.address)

            def say(text):
                saxo.send("PRIVMSG", self.sender, text)
            self.say = say

            def reply(text):
                saxo.send("PRIVMSG", self.sender, self.nick + ": " + text)
            self.reply = reply

            private = self.config.get("private")
            if self.private:
                if private == "deny":
                    self.blocked = True
            elif private == "only":
                self.blocked = True

            self.cmd = None

            cmdpfx = self.config["prefix"]
            if self.text.startswith(cmdpfx):
                text = self.text[len(cmdpfx):]
                if " " in text:
                    cmd, arg = text.split(" ", 1)
                else:
                    cmd, arg = text, ""
                self.cmd = cmd
                self.arg = arg

    def authorised(self):
        import re

        config_owner = self.config.get("owner", "")
        test_identity = False
        if not "!" in config_owner:
            test_identity = True
            config_owner = config_owner + "!*@*"

        mask = lambda g: "^" + re.escape(g).replace("\\*", ".*") + "$"
        matches = re.match(mask(config_owner), self.prefix) is not None
        if test_identity:
            return matches and self.identified
        return matches

    def client(self, *args):
        incoming.put(args)

# threaded
def socket_receive(sock):
    def receive_loop(sock, incoming):
        with sock.makefile("rb") as s:
            # TODO: How do we know when we're connected?
            incoming.put(("receiving",))
            for octets in s:
                incoming.put(("remote", octets))

    try: receive_loop(sock, incoming)
    except Exception as err:
        # Usually IOError, EOFError, socket.error, or ssl.SSLError
        debug(str(err))
    incoming.put(("disco_receiving",))

def flood_protection(recent):
    recent.append(time.monotonic())
    if len(recent) > 3:
        recent.popleft()
        if (recent[2] - recent[0]) < 2:
            # debug("sleeping 1")
            time.sleep(1)
        elif (recent[2] - recent[1]) < 1:
            # debug("sleeping 0.5")
            time.sleep(0.5)
        # else:
        #     debug("sleeping 0")
    else:
        # debug("sleeping 0.25")
        time.sleep(0.25)

# threaded
def socket_send(sock, flood=False):
    def sending(sock, flood=False):
        recent = collections.deque()
        with sock.makefile("wb") as s:
            incoming.put(("sending",))
            while True:
                octets = outgoing.get()
                if octets is None:
                    debug("Sending Thread: Requested quit")
                    return True

                debug("->", repr(octets.decode("utf-8", "replace")))
                s.write(octets)
                s.flush()

                if not flood:
                    flood_protection(recent)
        # TODO: Surely this is never reached?
        debug("Sending Thread: No more data")
        return False

    try: sending(sock, flood)
    except Exception as err:
        # Usually BrokenPipeError
        debug("Sending Thread: Error:", err)
    incoming.put(("disco_sending",))

class SaxoConnectionError(Exception):
    ...

regex_link = re.compile(r"(http[s]?://[^<> \"\x01]+)[,.]?")

def command_path(base, cmd):
    if ("\x00" in cmd) or (os.sep in cmd) or ("." in cmd):
        return None
    path = os.path.join(base, "commands", cmd)
    if (not os.path.isfile(path)) or (not os.path.getsize(path)):
        return None
    return path

def utf8(obj):
    return str(obj).encode("utf-8", "replace")

def utf8dict(data):
    return {utf8(key): utf8(value) for key, value in data.items()}

def process(env, cmd, path, arg):
    path = utf8(path)
    env = utf8dict(env)
    octets = utf8(arg)

    try: proc = subprocess.Popen([path, octets], env=env,
        stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    except PermissionError:
        outs = "The command file does not have executable permissions"
    except FileNotFoundError:
        # Might have been removed just after running this thread
        return
    else:
        authorised = "SAXO_AUTHORISED" in env
        private = not env.get("SAXO_SENDER", "#").startswith("#")
        timeout = 36 if (authorised and private) else 12

        try: outs, errs = proc.communicate(octets + b"\n", timeout=timeout)
        except subprocess.TimeoutExpired:
            proc.kill()
            # TODO: Use actual prefix
            outs = "Sorry, %s took too long" % cmd
        else:
            outs = outs.decode("utf-8", "replace")
            if "\n" in outs:
                outs = outs.splitlines()[0]

        code = proc.returncode or 0
        # Otherwise: TypeError: unorderable types: NoneType() > int()
        if (code > 0) and (not outs):
            # TODO: Use actual prefix
            outs = "Sorry, %s responded with an error" % cmd
    return outs

class Saxo(object):
    def __init__(self, base, opt):
        self.base = base
        self.opt = opt
        self.events = {}
        self.address = None
        self.discotimer = None
        self.receiving = False
        self.sending = False
        self.receiving_thread = None
        self.sending_thread = None
        self.reconnecting = False
        self.links = {}

        self.environment_cache = os.environ.copy()
        self.environment_cache["PYTHONPATH"] = saxo_path
        # TODO: This needs to be changed when setting nick
        self.environment_cache["SAXO_BOT"] = opt["client"]["nick"]
        self.environment_cache["SAXO_BASE"] = base
        self.environment_cache["SAXO_COMMANDS"] = \
            os.path.join(base, "commands")
        self.environment_cache["SAXO_VERSION"] = saxo_version

        self.config_cache = {}
        client_options = {
            "channels", # Channels to join on startup
            "nick", # Nickname of the bot, variable
            "owner", # Full address of the owner
            "prefix", # Command prefix
            "flood", # Whether or not to flood
            "private" # Whether to respond in private
        }

        for option in opt["client"]:
            if option in client_options:
                self.config_cache[option] = opt["client"].get(option)
            else:
                debug("Unknown option: %s" % option)

        for section in opt:
            if section == "client":
                continue
            self.config_cache[section] = dict(opt[section])

    def run(self):
        self.load()
        self.connect()
        self.handle()

    def load(self):
        # Update symlinks
        common.populate(saxo_path, self.base)

        # Load events
        first = not self.events
        self.events.clear()

        def module_exists(name):
            try: imp.find_module(name)
            except ImportError:
                return False
            else: return True

        if first and module_exists("plugins"):
            debug("Warning: a 'plugins' module already exists")

        if first and ("plugins" in sys.modules):
            raise ImportError("'plugins' duplicated")

        # This means we're using plugins as a namespace module
        # Might have to move saxo.path's plugins/ to something else
        # Otherwise it gets unionised into the namespace module
        # if self.base not in sys.path:
        # - not needed, because we clear up below
        sys.path[:0] = [self.base]

        plugins = os.path.join(self.base, "plugins")
        plugins_package = importlib.__import__("plugins")
        if next(iter(plugins_package.__path__)) != plugins:
            # This is very unlikely to happen, because we pushed self.base
            # to the front of sys.path, but perhaps some site configuration
            # or other import mechanism may affect this
            raise ImportError("non-saxo 'plugins' module")

        setups = {}
        for name in os.listdir(plugins):
            if ("_" in name) or (not name.endswith(".py")):
                continue

            name = "plugins." + name[:-3]
            if not name in sys.modules:
                try: module = importlib.import_module(name)
                except Exception as err:
                    debug("Error loading %s:" % name, err)
            elif first:
                raise ImportError("%r duplicated" % name)
            else:
                module = sys.modules[name]
                try: module = imp.reload(module)
                except Exception as err:
                    debug("Error reloading %s:" % name, err)

            for attr in dir(module):
                obj = getattr(module, attr)

                if hasattr(obj, "saxo_event"):
                    try: self.events[obj.saxo_event].append(obj)
                    except KeyError:
                        self.events[obj.saxo_event] = [obj]

                elif hasattr(obj, "saxo_setup"):
                    obj.saxo_name = module.__name__ + "." + obj.__name__
                    setups[obj.saxo_name] = obj

            # debug("Loaded module:", name)

        debug("%s setup functions" % len(setups))

        graph = {}
        for setup in setups.values():
            deps = ["plugins." + dep for dep in setup.saxo_deps]
            graph[setup.saxo_name] = deps

        database_filename = os.path.join(self.base, "database.sqlite3")
        with sqlite.Database(database_filename) as self.db:
            for name in common.tsort(graph):
                debug(name)
                if name in setups:
                    setups[name](self)
                else:
                    debug("Warning: Missing dependency:", name)

        sys.path[:1] = []

    def connect(self):
        try: self.connect_sock()
        except Exception as err:
           raise SaxoConnectionError(str(err))

        self.first = True
        # TODO: Reset other state? e.g. self.address

        receiving = (socket_receive, self.sock)
        self.receiving_thread = common.thread(*receiving)

        sending = (socket_send, self.sock, "flood" in self.opt["client"])
        self.sending_thread = common.thread(*sending)

    def connect_sock(self):
        host = self.opt["server"]["host"]
        port = int(self.opt["server"]["port"])

        debug("Connecting to %s:%s" % (host, port))
        # self.sock.connect((host, port))
        self.sock = socket.create_connection((host, port))
        # self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if "ssl" in self.opt["server"]:
            import ssl
            debug("Warning: Using SSL, but not validating the cert!")
            self.sock = ssl.wrap_socket(
                self.sock,
                server_side=False,
                cert_reqs=ssl.CERT_NONE) # TODO: or CERT_REQUIRED

    def disconnect(self):
        # NOTE: *Do not* close the sending thread gracefully
        # The socket shutdown code below will be invoked before the queue
        # item reaches the socket thread. That means the *next* thread will
        # pick it up, because it'll quit with a pipe error anyway

        # Close the socket forcefully
        try: self.sock.shutdown(socket.SHUT_RDWR)
        except: ...

        try: self.sock.close()
        except: ...

        self.cancel_discotimer()

    def socket_threads_active(self):
        receiving = self.receiving_thread.is_alive()
        sending = self.sending_thread.is_alive()
        debug("RECV, SEND:", receiving, sending)
        return receiving or sending

    def cancel_discotimer(self):
        if self.discotimer is not None:
            try:
                was_finished = self.discotimer.finished.is_set()
                # Note: .cancel() still works on an expired timer!
                # It just calls .finished.set()
                self.discotimer.cancel()
                if not was_finished:
                    debug("Cancelled probably unfinished disco timer")
                self.discotimer = None
            except:
                ...

    def handle(self):
        while True:
            instruction_args = incoming.get()
            instruction = instruction_args[0]
            args = tuple(instruction_args[1:])

            if instruction not in {"instances", "remote"}:
                debug("handle:", instruction, args)

            if not isinstance(instruction, str):
                continue

            method_name = "instruction_" + instruction
            if hasattr(self, method_name):
                method = getattr(self, method_name)
                try: method(*args)
                except Exception as err:
                    debug("handle error:", err)
                    # raise err # - for debug
            else:
                debug("Unknown instruction:", instruction)

    def instruction_address(self, address):
        self.address = address

    def instruction_connect(self, delay=5):
        # Warning: This quits the bot if the bot is already connected!
        # Use instruction_reconnect if you want to restart the bot

        # NOTE: The instruction mechanism is synchronous
        # This means we can't use self.receiving etc. to check connectivity
        # We rely on self.disconnect() to have done the right thing
        # The following is a check to make sure that it has
        if self.socket_threads_active():
            # Wait up to six seconds for the threads to quit
            for attempt in range(12):
                time.sleep(0.5)
                if not self.socket_threads_active():
                    break
            else:
                # TODO: If the threads are still active, they should be killed
                # Unfortunately, threads in python can't be killed
                debug("ERROR! Unable to stop the socket threads")
                exit(1)

        if "flood" not in self.opt["client"]:
            time.sleep(3)

        try: self.connect()
        except SaxoConnectionError as err:
            # Retry, with 1sec more delay, up to a maximum of 30sec delay
            def connect():
                incoming.put(("connect", min(30, delay + 1)))
            t = threading.Timer(delay, connect)
            t.start()

    def instruction_connected(self):
        if ":connected" in self.events:
            msg = Message(self, b"NOOP")
            for function in self.events[":connected"]:
                function(msg)

    def instruction_command(self, prefix, sender, cmd, arg):
        self.command(prefix, sender, cmd, arg)

    def instruction_disco_receiving(self):
        self.receiving = False
        debug("Sending disconnected event to scheduler")
        scheduler.incoming.put(("disconnected", ()))
        if self.sending:
            outgoing.put(None)
        else:
            incoming.put(("connect",))

    def instruction_disco_sending(self):
        self.sending = False
        if self.receiving:
            self.disconnect()
        else:
            incoming.put(("connect",))

    def instruction_instances(self):
        our_pid = os.getpid()
        database_filename = os.path.join(self.base, "database.sqlite3")
        with sqlite.Database(database_filename) as db:
            pids = [row[0] for row in db["saxo_instances"]]
        if len(pids) == 1:
            if pids[0] == our_pid:
                return

        debug("Found another saxo instance! %s" % pids)
        self.send("QUIT", "Another saxo instance was detected")
        self.disconnect()
        exit(0)

    def instruction_join(self, channel):
        # NOTE: .visit can still be followed by .join
        # The JOIN will just fail on the server

        channels = self.opt["client"]["channels"].split(" ")
        if channel in channels:
            # TODO: .visit it back?
            return
        channels.append(channel)
        channels = " ".join(channels)

        self.update_config("client", "channels", channels)
        self.send("JOIN", channel)

    def instruction_link(self, channel, link):
        self.links[channel] = link

    def instruction_message(self, text):
        debug("IPC message:", text)

    def instruction_msg(self, destination, text):
        self.send("PRIVMSG", destination, text)

    def instruction_noop(self):
        ...

    def instruction_part(self, channel):
        # NOTE: .leave can still be followed by .part
        # The PART will just fail on the server

        channels = self.opt["client"]["channels"].split(" ")
        if channel not in channels:
            # TODO: .leave it anyway?
            return
        channels.remove(channel)
        channels = " ".join(channels)

        self.update_config("client", "channels", channels)
        self.send("PART", channel)

    def instruction_periodic(self, name, period, cmd, arg, sender=None):
        database_filename = os.path.join(self.base, "database.sqlite3")
        with sqlite.Database(database_filename) as db:
            p = (name, period, int(time.time()), b"scheduled",
                 common.b64pickle((cmd, arg, sender)))
            # TODO: This fails silently if there's a type error?
            db["saxo_periodic"].replace(p)

    def instruction_ping(self):
        now = time.time()

        # Don't ping if connected less than a minute ago
        if self.receiving > (now - 60):
            debug("Won't start discotimer, because we recently reconnected")
            return

        self.send("PING", self.opt["client"]["nick"])

        def reconnect():
            incoming.put(("reconnect",))

        # Make sure the timer is shorter than the pingloop!
        # Default pingloop period is 180
        self.discotimer = threading.Timer(30, reconnect)
        self.discotimer.start()

    def instruction_prefix(self, pfx):
        self.update_config("client", "prefix", pfx)

    def instruction_quit(self):
        # Never call this from a thread, otherwise this can give an OSError
        # TODO: Get the sender to pick this up and disconnect from there?
        # Could be a problem if the sender has broken
        self.send("QUIT")
        self.disconnect()
        exit(0)

    def instruction_receiving(self):
        self.receiving = time.time()

        scheduler.incoming.put(("connected", ()))
        def start_scheduler():
            scheduler.incoming.put(("start", ()))
        start = threading.Timer(3, start_scheduler)
        start.start()

        # TODO: Check that we really are connected
        # TODO: Unit test for :connected event
        incoming.put(("connected",))

    def instruction_reconnect(self):
        # disco_* will automatically reconnect
        if self.receiving or self.receiving_thread.is_alive():
            # Close the socket, which forces the receiving thread to quit
            self.disconnect()
        elif self.sending or self.sending_thread.is_alive():
            # Signal the sending thread to quit
            outgoing.put(None)
        else:
            debug("Reconnect requested during reconnect")

    def instruction_reload(self, destination=None):
        before = time.time()
        self.load()
        elapsed = time.time() - before
        if destination:
            self.send("PRIVMSG", destination,
                "Reloaded in %s seconds" % round(elapsed, 3))

    def instruction_remote(self, octets):
        debug(repr(octets))
        msg = Message(self, octets)

        if msg.command == "PRIVMSG":
            if msg.blocked:
                return

            if msg.cmd is not None:
                self.command(msg)

        elif msg.command == "PONG":
            self.cancel_discotimer()

        def run(name, msg):
            if name in self.events:
                for function in self.events[name]:
                    try: function(msg)
                    except Exception as err:
                        debug("Error:", function.__name__ + ":", err)

        if self.first is True:
            run(":1st", msg)
            self.first = False

        run(msg.command, msg)
        run("*", msg)

    def instruction_schedule(self, unixtime, command, args):
        command = command.encode("ascii")
        args = common.b64pickle(args)
        # TODO: Why not just add it to the database ourselves?
        scheduler.incoming.put(("schedule.add", (unixtime, command, args)))

    def instruction_scheduled(self, cmd, arg, sender=None):
        self.scheduled_command(cmd, arg, sender=sender)

    def instruction_send(self, *args):
        self.send(*args)

    def instruction_sending(self):
        self.sending = True

    def command(self, msg):
        cmd, arg = msg.cmd, msg.arg
        path = command_path(self.base, cmd)
        if path is None:
            return

        def command_process(env, cmd, path, arg):
            outs = process(env, cmd, path, arg)
            if outs:
                self.send("PRIVMSG", msg.sender, outs)

        env = self.environment_cache.copy()
        env["SAXO_NICK"] = msg.nick
        env["SAXO_SENDER"] = msg.sender
        if msg.sender in self.links:
            env["SAXO_URL"] = self.links[msg.sender]
        if msg.authorised():
            env["SAXO_AUTHORISED"] = "1"
        common.thread(command_process, env, cmd, path, arg)

    def scheduled_command(self, cmd, arg, sender=None):
        path = command_path(self.base, cmd)
        if path is None:
            return

        def command_process(env, cmd, path, arg):
            outs = process(env, cmd, path, arg)
            if outs and (sender is not None):
                self.send("PRIVMSG", sender, outs)

        env = self.environment_cache.copy()
        env["SAXO_SCHEDULED"] = "1"
        common.thread(command_process, env, cmd, path, arg)

    def update_config(self, section, option, value):
        self.opt[section][option] = value
        config = os.path.join(self.base, "config")
        with open(config, "w", encoding="utf-8") as f:
            self.opt.write(f)
        if section == "client":
            self.config_cache[option] = value
        else:
            self.config_cache[section][option] = value
        # TODO: Change the database

    def send(self, *args):
        # TODO: Loop detection
        if len(args) > 1:
            args = args[:-1] + (":" + args[-1],)
        text = re.sub(r"[\r\n]", "", " ".join(args))

        if len(args) == 3:
            if (args[0] == "PRIVMSG") and args[1].startswith("#"):
                search = regex_link.search(args[2])
                if search:
                    incoming.put(("link", args[1], search.group(1)))

        outgoing.put(text.encode("utf-8", "replace")[:510] + b"\r\n")

def serve(sockname, incoming):
    if os.path.exists(sockname):
        os.remove(sockname)

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind(sockname)
    sock.listen(1)

    def listen(sock):
        while True:
            connection, client = sock.accept()
            def handle(connection, client):
                try:
                    for octets in connection.makefile("rb"):
                        try:
                            text = octets.decode("ascii", "replace")
                            text = text.strip("\n")

                            if " " in text:
                                instruction, data = text.split(" ", 1)
                                args = common.b64unpickle(data)
                            else:
                                instruction, args = text, tuple()

                            incoming.put((instruction,) + args)
                        except Exception as err:
                            debug("ERROR!", err.__class__.__name__, err)
                finally:
                    connection.close()
            common.thread(handle, connection, client)
    common.thread(listen, sock)

E_NO_CONFIG = """
Are you sure this is a saxo configuration directory? If you need to make a new
configuration directory, use the `saxo create` command.
"""

def start(base):
    # TODO: Check when two clients are running
    common.exit_cleanly()
    # http://stackoverflow.com/questions/11423225
    # IGN rather than DFL, otherwise Popen.communicate can quit saxo
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)

    opt = configparser.ConfigParser(interpolation=None)
    config = os.path.join(base, "config")
    if not os.path.isfile(config):
        error("missing config file in: `%s`" % config, E_NO_CONFIG)
    opt.read(config)
    # TODO: Defaulting?
    # TODO: Warn if the config file is widely readable?

    sockname =  os.path.join(base, "client.sock")
    serve(sockname, incoming)
    os.chmod(sockname, 0o600)

    # NOTE: If using os._exit, this doesn't work
    def remove_sock(sockname):
        if os.path.exists(sockname):
            os.remove(sockname)
    atexit.register(remove_sock, sockname)

    sched = scheduler.Scheduler(incoming)
    common.thread(sched.start, base)

    saxo = Saxo(base, opt)
    saxo.run()



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import html.entities
import re
import urllib.parse
import urllib.request

# NOTE: (?i) does work in byte instances
_regex_charset = re.compile(b"(?i)<meta[^>]+charset=[\"']?([^\"'> \r\n\t]+)")
_regex_entity = re.compile(r"&([^;\s]+);")
_regex_key = re.compile(r'([^=]+)')
_regex_value = re.compile(r'("[^"\\]*(?:\\.[^"\\]*)*"|[^;]+)')

user_agent = "Mozilla/5.0 (Services)"
# modern_user_agent = " ".join([
#     "Mozilla/5.0",
#     "(Macintosh; Intel Mac OS X 10.9; rv:26.0)"
#     "Gecko/20100101 Firefox/26.0"])
modern_user_agent = " ".join([
    "Mozilla/5.0",
    "(Macintosh; Intel Mac OS X 10.10; rv:37.0)"
    "Gecko/20100101 Firefox/37.0"])

def content_type(info):
    mime = None
    encoding = None

    def parse(parameters):
        while parameters:
            match = _regex_key.match(parameters)
            if not match:
                break

            key = match.group(1)
            parameters = parameters[len(key):]
            if parameters.startswith("="):
                parameters = parameters[1:]

            match = _regex_value.match(parameters)
            if not match:
                break

            value = match.group(1)
            if value.startswith('"'):
                value = value[1:-1].replace('\\"', '"')
            parameters = parameters[len(value):]

            if parameters.startswith(";"):
                parameters = parameters[1:]

            key = key.lower().strip(" \t")
            value = value.lower().strip(" \t")
            yield key, value

    if "Content-Type" in info:
        header = info["Content-Type"]
        if ";" in header:
            mime, parameters = header.split(";", 1)
        else:
            mime, parameters = header, ""

        for key, value in parse(parameters):
            if key == "charset":
                encoding = value
                break
    return mime, encoding

def decode_entities(hypertext):
    def entity(match):
        name = match.group(1).lower()

        if name.startswith("#x"):
            return chr(int(name[2:], 16))
        elif name.startswith("#"):
            return chr(int(name[1:]))
        elif name in html.entities.name2codepoint:
            return chr(html.entities.name2codepoint[name])
        return "[" + name + "]"

    def default(match):
        try: return entity(match)
        except: return match.group(1)
    return _regex_entity.sub(default, hypertext)

def construct(url, query=None):
    safe = "".join(chr(i) for i in range(0x01, 0x80))
    base = urllib.parse.quote(url, safe=safe)
    if query:
        query = urllib.parse.urlencode(query)
        return "?".join((base, query))
    return base

def request(url, query=None, data=None, method="GET",
        limit=None, follow=True, headers=None, modern=False):
    headers = {} if (headers is None) else headers

    response = {}
    if "User-Agent" not in headers:
        modern = modern is True
        headers["User-Agent"] = modern_user_agent if modern else user_agent
    response["request-headers"] = headers

    safe = "".join(chr(i) for i in range(0x01, 0x80))
    base = urllib.parse.quote(url, safe=safe)
    if query:
        query = urllib.parse.urlencode(query)
        response["request-url"] = "?".join((base, query))
    else:
        response["request-url"] = base

    class ErrorHandler(urllib.request.HTTPDefaultErrorHandler):
        def http_error_default(self, req, fp, code, msg, hdrs):
            return fp

    handlers = [ErrorHandler()]
    if follow:
        handlers.append(urllib.request.HTTPRedirectHandler())

    opener = urllib.request.build_opener(*handlers)
    urllib.request.install_opener(opener)

    params = {
        "url": response["request-url"],
        "headers": response["request-headers"]
    }

    if data is not None:
        if isinstance(data, dict):
            data = urllib.parse.urlencode(data)
            params["data"] = data.encode("utf-8", "replace")
        elif isinstance(data, bytes):
            params["data"] = data
        elif isinstance(data, str):
            params["data"] = data.encode("utf-8", "replace")
        else:
            raise Exception("Unknown data type: %s" % type(data))

    req = urllib.request.Request(**params)
    with urllib.request.urlopen(req) as res:
        response["url"] = res.url
        response["status"] = res.status # int
        response["info"] = res.info()
        response["headers"] = {
            a.lower(): b for (a, b) in response["info"].items()
        }

        if method in {"GET", "POST"}:
            if limit is None:
                response["octets"] = res.read()
            else:
                response["octets"] = res.read(limit)

    if "Content-Encoding" in response["info"]:
        if response["info"]["Content-Encoding"] == "gzip":
            from gzip import GzipFile
            from io import BytesIO
            sio = BytesIO(response["octets"])
            gz = GzipFile(fileobj=sio)
            try: response["octets"] = gz.read()
            except OSError:
                # e.g. not gzip encoded, despite the site saying it is
                ...

    mime, encoding = content_type(response["info"])
    if mime:
        response["mime"] = mime
    if encoding:
        response["encoding"] = encoding
        response["encoding-source"] = "Content-Type"

    if mime and ("octets" in response):
        if ("/html" in mime) or ("/xhtml" in mime):
            search = _regex_charset.search(response["octets"])
            if search:
                html_encoding = search.group(1).lower()
                html_encoding = html_encoding.decode("ascii", "replace")

                if encoding and (encoding == html_encoding):
                        response["encoding-source"] += ", HTML"
                else: # TODO: Precedence check
                    response["encoding"] = html_encoding
                    response["encoding-source"] = "HTML"

    if "octets" in response:
        def guess_encoding(response):
            try: response["text"] = response["octets"].decode("utf-8")
            except UnicodeDecodeError:
                response["text"] = response["octets"].decode("iso-8859-1")
                response["encoding"] = "iso-8859-1"
            else:
                response["encoding"] = "utf-8"
            response["encoding-source"] = "heuristic"

        if "encoding" in response:
            encoding = response["encoding"]
            try: response["text"] = response["octets"].decode(encoding)
            except (UnicodeDecodeError, LookupError):
                guess_encoding(response)
        else:
            guess_encoding(response)

    if mime and ("text" in response):
        if ("/html" in mime) or ("/xhtml" in mime):
                response["html"] = response["text"]
                response["text"] = decode_entities(response["text"])
                response["decoded-entities"] = True

    return response

def lynx(url, query=None, data=None, source=True):
    import subprocess
    url = construct(url, query=query)
    command = ["lynx"]
    if data is not None:
        command.append("--post_data")
    command.append("--dump")
    if source is True:
        command.append("--source")
    # command.append("--accept_all_cookies")
    command.append(url)
    try: octets = subprocess.check_output(command, input=data)
    except FileNotFoundError as err:
        return "LynxNotFound"
    except Exception as err:
        return "Error: %s" % err
    return octets.decode("utf-8", "ignore")



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import saxo

@saxo.event("251")
def nickserv251(irc):
    if "plugins" in irc.config:
        if "nickserv" in irc.config["plugins"]:
            # Works on freenode, not sure about other servers. Easy to modify
            nick = irc.config["nick"]
            password = irc.config["plugins"]["nickserv"]
            irc.msg("NickServ", "IDENTIFY %s %s" % (nick, password))

@saxo.event("451")
def nickserv451(irc):
    if "plugins" in irc.config:
        if "nickserv" in irc.config["plugins"]:
            # Works on freenode, not sure about other servers. Easy to modify
            nick = irc.config["nick"]
            password = irc.config["plugins"]["nickserv"]
            irc.msg("NickServ", "IDENTIFY %s %s" % (nick, password))



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import saxo

@saxo.setup
def instances(irc):
    if not "saxo_instances" in irc.db:
        irc.db["saxo_instances"].create(
            ("pid", int))

@saxo.setup
def periodic(irc):
    sqlite3_schema = [(0, 'name', 'TEXT', 0, None, 1),
                      (1, 'period', 'INTEGER', 0, None, 0),
                      (2, 'recent', 'INTEGER', 0, None, 0),
                      (3, 'command', 'BLOB', 0, None, 0),
                      (4, 'args', 'BLOB', 0, None, 0)]

    if "saxo_periodic" in irc.db:
        schema = list(irc.db["saxo_periodic"].schema())
        if schema != sqlite3_schema:
            del irc.db["saxo_periodic"]

    if "saxo_periodic" not in irc.db:
        irc.db["saxo_periodic"].create(("name", "TEXT PRIMARY KEY"),
                                       ("period", int),
                                       ("recent", int),
                                       ("command", bytes),
                                       ("args", bytes))

@saxo.setup
def schedule(irc):
    if not "saxo_schedule" in irc.db:
        irc.db["saxo_schedule"].create(
            ("unixtime", int),
            ("command", bytes),
            ("args", bytes))



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import os
import saxo

def dependencies(*deps):
    def decorator(function):
        function.saxo_deps = deps
        return function
    return decorator

@saxo.setup
@dependencies("schema.instances")
def populate(irc):
    # Remove any values from previous instances
    for row in irc.db["saxo_instances"]:
        del irc.db["saxo_instances"][row]

    # Add our value
    irc.db["saxo_instances"].insert((os.getpid(),))



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import atexit
import os
import signal
import sys

def writeable(path):
    if os.path.exists(path):
        return os.access(path, os.W_OK)

    directory = os.path.dirname(path)
    if not os.path.isdir(directory):
        return False

    return os.access(directory, os.W_OK)

def fork(n):
    try: pid = os.fork()
    except OSError as err:
        print("Error: Unable to fork on this OS: %s" % err)
        print("Use the --foreground option to avoid running as a daemon")
        sys.exit(1)
    else:
        if pid > 0:
            sys.exit(0)

def redirect(a, b):
    os.dup2(b.fileno(), a.fileno())

def start(pidfile, output):
    if not writeable(pidfile):
        print("Error: Can't write to PID file: " + str(pidfile))
        sys.exit(1)

    fork(1)

    os.chdir("/")
    os.setsid()
    os.umask(0)

    fork(2)

    pid = os.getpid()

    print("Running saxo as PID %s" % pid)
    print("This PID is also saved in %s" % pidfile)

    sys.stdout.flush()
    sys.stderr.flush()

    redirect(sys.stdin, open(os.devnull, "r"))
    redirect(sys.stdout, output)
    redirect(sys.stderr, output)

    with open(pidfile, "w") as f:
        f.write(str(pid) + "\n")

    def delete_pidfile():
        if os.path.isfile(pidfile):
            os.remove(pidfile)
    atexit.register(delete_pidfile)

    signal.signal(signal.SIGHUP, signal.SIG_IGN)

    return pid



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import re
import sqlite3

class Table(object):
    def __init__(self, connection, name):
        self.connection = connection
        self.name = name

    def __iter__(self):
        return self.rows()

    def __delitem__(self, row):
        fields = []
        for field in self.schema():
            fields.append(field[1])

        if len(row) == len(fields):
            query = "DELETE FROM %s WHERE " % self.name
            query += " AND ".join(["%s=?" % field for field in fields])
            cursor = self.connection.cursor()
            cursor.execute(query, row)
            self.connection.commit()
        else:
            raise ValueError("Wrong length: %s" % row)

    def create(self, *schema):
        cursor = self.connection.cursor()
        types = {
            None: "NULL",
            int: "INTEGER",
            float: "REAL",
            str: "TEXT",
            bytes: "BLOB"
        }
        schema = ", ".join(a + " " + types.get(b, b) for (a, b) in schema)
        query = "CREATE TABLE IF NOT EXISTS %s (%s)" % (self.name, schema)
        cursor.execute(query)
        cursor.close()

    def insert(self, row, *rows, commit=True):
        cursor = self.connection.cursor()
        size = len(row)
        args = ",".join(["?"] * size)
        query = "INSERT INTO %s VALUES(%s)" % (self.name, args)

        cursor.execute(query, tuple(row))
        for extra in rows:
            cursor.execute(query, tuple(extra))

        if commit:
            self.connection.commit()
        cursor.close()

    def replace(self, row, *rows, commit=True):
        # TODO: command="INSERT", then command="INSERT OR REPLACE"
        cursor = self.connection.cursor()
        size = len(row)
        args = ",".join(["?"] * size)
        query = "INSERT OR REPLACE INTO %s VALUES(%s)" % (self.name, args)

        cursor.execute(query, tuple(row))
        for extra in rows:
            cursor.execute(query, tuple(extra))

        if commit:
            self.connection.commit()
        cursor.close()

    def rows(self, order=None):
        cursor = self.connection.cursor()
        query = "SELECT * FROM %s" % self.name

        if isinstance(order, str):
            if order.isalpha():
                query += " ORDER BY %s" % order

        cursor.execute(query)
        while True:
            result = cursor.fetchone()
            if result is None:
                break
            yield result
        cursor.close()

    def schema(self):
        cursor = self.connection.cursor()
        query = "PRAGMA table_info(%s)" % self.name
        cursor.execute(query)
        while True:
            result = cursor.fetchone()
            if result is None:
                break
            yield result
        cursor.close()

class Database(object):
    def __init__(self, path):
        self.path = path
        self.connection = sqlite3.connect(path)

        def regexp(pattern, text):
            return re.search(pattern, text) is not None
        self.connection.create_function("REGEXP", 2, regexp)

    def __iter__(self):
        raise NotImplemented

    def __delitem__(self, key):
        if key in self:
            query = "DROP TABLE %s" % key
            cursor = self.connection.cursor()
            cursor.execute(query)
            cursor.close()

    def __contains__(self, key):
        cursor = self.connection.cursor()
        query = "SELECT name FROM sqlite_master WHERE type='table' AND name=?"
        cursor.execute(query, (key,))
        result = cursor.fetchone() is not None
        cursor.close()
        return result

    def __getitem__(self, key):
        return Table(self.connection, key)

    def __enter__(self, *args, **kargs):
        return self

    def __exit__(self, *args, **kargs):
        # TODO: Check for changes to commit?
        # self.connection.commit()
        self.connection.close()

    def commit(self):
        self.connection.commit()

    def execute(self, text, *args):
        cursor = self.connection.cursor()
        if not args:
            cursor.execute(text)
        else:
            cursor.execute(text, args)
        return cursor

    def query(self, text, *args):
        cursor = self.execute(text, *args)
        # Duplicate rows are sometimes given,
        # even when sqlite3 was compiled thread-safe
        previous = None
        while True:
            result = cursor.fetchone()
            if result is None:
                break
            if result == previous:
                continue
            yield result
            previous = result
        cursor.close()

def test():
    import os

    filename = "/tmp/saxo-test.sqlite3"
    if os.path.isfile(filename):
        os.remove(filename)

    with Database(filename) as db:
        assert "example" not in db
        db["example"].create(
            ("name", str),
            ("size", int))

        assert "example" in db
        db["example"].insert(
            ("pqr", 5),
            ("abc", 10))

        print(list(db["example"]))
        assert list(db["example"].rows(order="name")) == [('abc', 10), ('pqr', 5)]
        assert list(db["example"].rows(order="size")) == [('pqr', 5), ('abc', 10)]
        print(list(db["example"].schema()))

        del db["example"][("pqr", 5)]
        print(list(db["example"]))

    os.remove(filename)

if __name__ == "__main__":
    test()



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import time
import saxo

def dependencies(*deps):
    def decorator(function):
        function.saxo_deps = deps
        return function
    return decorator

def replace(irc, *args):
    irc.db["saxo_periodic"].replace(args)

@saxo.setup
@dependencies("schema.periodic")
def populate(irc):
    # Periodic tasks whose names start with "@" are temporary
    # This code deletes those tasks on initialisation
    temporary = []
    for row in irc.db["saxo_periodic"].rows():
        if row[0].startswith("@"):
            temporary.append(row)
    for row in temporary:
        del irc.db["saxo_periodic"][row]

    current = int(time.time())
    # TODO: Or "check_connection" instead of "ping"
    replace(irc, "check connection", 180, current, b"ping", b"")
    replace(irc, "check unique", 20, current, b"instances", b"")



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

# TODO list:
# - Think about folding common.py into saxo.py
# - Allow wildcards in [client].owner
# - Implement [plugins].sync, copy or symlink as options
# - Document the saxo, irc, self, and instruction interfaces
# - Think about a packaging system
# - Maybe write up a little guide to installing py3.3 with sqlite support
# - Allow JSON submission to the socket IPC interface
# - Write docstrings for all the public interfaces
# - Allow multiple, possibly configurable, lines of output from commands?
# - Plugins for pre_exec?
# - Make it possible to set periodic commands
# - Make sure the test server exits correctly
# - Dump a copy of the initialised config to the database
# - Document the database tables
# - Delete and select methods for sqlite.Table

import os
import sys

from collections import namedtuple

if sys.version_info < (3, 3):
    raise Exception("Requires Python 3.3 or later")

path = None

if "__file__" in vars():
    if __file__:
        path = os.path.abspath(__file__)
        path = os.path.realpath(path)
        path = os.path.dirname(path)

if "__path__" in vars():
    if __path__:
        for directory in __path__:
            if path is None:
                path = directory
            elif path != directory:
                raise Exception("Can't create saxo.path")

if path is None:
   raise Exception("Can't create saxo.path")

with open(os.path.join(path, "version")) as f:
    version = f.read().rstrip("\r\n")

# Move this to saxo.pipe?
saxo_version = os.environ.get("SAXO_VERSION")
if (saxo_version or version) != version:
    message = "Error: Version mismatch: Saxo %s attempted to import Saxo %s"
    msg = message % (version, saxo_version)
    print(msg)
    raise ImportError(msg)

Version = namedtuple("Version", ["major", "minor", "serial"])
version_info = Version(*(int(n) for n in version.split(".")))

__all__ = ["path", "version", "version_info"]

# Decorators:
#
# saxo.command(*args, *, authorised=False, owner=False, private=False)
# saxo.event(command="*")
# saxo.pipe(function)
# saxo.setup(function)
#
# Other:
#
# saxo.client(command, *args, base=None)
# saxo.database(name=None)
# saxo.env(name)
# saxo.request(*args, **kargs)
# saxo.script(argv)

# TODO: environment modification?

def public(function):
    __all__.append(function.__name__)
    return function

@public
def client(command, *args, base=None):
    import base64
    import pickle
    import socket

    if base is None:
        base = os.environ["SAXO_BASE"]

    sockname = os.path.join(base, "client.sock")
    client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    client.connect(sockname)

    command = command.encode("ascii", "replace")
    pickled = pickle.dumps(args)
    client.send(command + b" " + base64.b64encode(pickled) + b"\n")
    client.close()

@public
def command(*args, authorised=False, owner=False, private=False):
    # TODO: arity, etc.
    def plugin_command(name, owner=False):
        print("Warning: Use of saxo.command is deprecated!")

        def decorate(function):
            @event("PRIVMSG")
            def inner(irc):
                prefix = irc.config["prefix"]
                if irc.text.startswith(prefix):
                    length = len(prefix)
                    text = irc.text[length:]

                    if " " in text:
                        cmd, arg = text.split(" ", 1)
                    else:
                        cmd, arg = text, ""

                    if cmd == name:
                        irc.arg = arg
                        if (not owner) or irc.authorised():
                            function(irc)
            return inner
        return decorate

    if args:
        if len(args) == 1:
            return plugin_command(args[0], owner=owner)
        elif (len(args) == 2) and (owner is False):
            return plugin_command(args[0], owner=args[1])
        else:
            raise ValueError("Incorrect use of saxo.command")
    elif owner is not False:
        raise ValueError("No such keyword argument: owner")

    def decorator(function):
        if authorised and (not env("authorised")):
            return
        if private and env("sender").startswith("#"):
            return
        pipe(function)
    return decorator

# saxo_event
# saxo_periodic
# saxo_private
# saxo_schedule
# saxo_seen
# saxo_setup
# saxo_to
# saxo_unicode

# TODO: command -> group?
@public
def data(name, value=None, *, command=None, check=True, delete=False):
    import __main__ as main

    # Save PEP 3122!
    if "." in __name__:
        from .sqlite import Database
    else:
        from sqlite import Database

    if hasattr(main, "__file__"):
        cmd = os.path.basename(main.__file__)
    else:
        cmd = "?"

    if command is None:
        command = cmd
    elif check is True:
        if command != cmd:
            raise ValueError("%r != %r" % (command, cmd))

    base = env("base")
    if not base:
        # or if base is None?
        raise ValueError("Missing SAXO_BASE")

    key = command + "." + name
    if value is not None:
        if not isinstance(value, str):
            raise ValueError("Expected value to be a str instance")
        if delete is True:
            raise ValueError("delete=True is incompatible with a value")

    path = os.path.join(base, "database.sqlite3")
    with Database(path) as db:
        if delete is True:
            query = "DELETE FROM saxo_data WHERE key = ?"
            db.execute(query, key)
            db.commit()
            return True
        elif value is not None:
            import sqlite3
            try: db["saxo_data"].replace((key, value))
            except sqlite3.OperationalError:
                db["saxo_data"].create(("key", "TEXT PRIMARY KEY"),
                                       ("value", str))
                db["saxo_data"].replace((key, value))
        else:
            query = "SELECT * FROM saxo_data where key = ?"
            # TODO: Table.__getitem__
            try: result = list(db.query(query, key))[0][1]
            except: result = None
            return result

@public
def database(name=None, dotdir=False):
    # Save PEP 3122!
    if "." in __name__:
        from .sqlite import Database
    else:
        from sqlite import Database

    if name is None:
        base = env("base")
        if base is None:
            if dotdir is True:
                base = os.path.expanduser("~/.saxo")
            else:
                raise ValueError("No SAXO_BASE found")
        name = os.path.join(base, "database.sqlite3")

    return Database(name)

# saxo.database.path?
# TODO: Does saxo.database().path work?

# saxo.env("base")
# saxo.env("bot")
# saxo.env("commands")
# saxo.env("nick")
# saxo.env("sender")
# saxo.env("url")

@public
def env(name, alternative=None):
    return os.environ.get("SAXO_%s" % name.upper(), alternative)

# TODO: priority?
@public
def event(command="*"):
    def decorate(function):
        function.saxo_event = command
        return function
    return decorate

@public
def lynx(*args, **kargs):
    # Save PEP 3122!
    if "." in __name__:
        from . import web
    else:
        import web
    return web.lynx(*args, **kargs)

@public
def pipe(function):
    # This gives you:
    # * concise code
    # * clean exiting
    # * SAXO_COMMANDS check and possible setup
    # * input surrogate decoding
    # * output encoding
    # * a custom error wrapper

    # TODO: Would like to run this in caller __name__ == "__main__"
    # __name__ here is "core"
    # try import __main__ as main; main.__name__?
    # Doesn't work. Would probably have to do:
    # @saxo.command(name=__name__)

    # Save PEP 3122!
    if "." in __name__:
        from . import common
    else:
        import common

    common.exit_cleanly()

    if not os.environ.get("SAXO_COMMANDS"):
        os.environ["SAXO_COMMANDS"] = os.path.dirname(sys.argv[0])

    if len(sys.argv) > 1:
        arg = " ".join(sys.argv[1:])
        # We have to do this because python converts arguments to surrogates
        # http://stackoverflow.com/a/7077803
        arg = os.fsencode(arg).decode("utf-8")
    elif not sys.stdin.isatty():
        arg = sys.stdin.buffer.readline()
        arg = arg.rstrip(b"\r\n")
        arg = arg.decode("utf-8")
    else:
        arg = ""

    try: result = function(arg)
    except Exception as err:
        import traceback

        python = err.__class__.__name__ + ": " + str(err)
        stack = traceback.extract_tb(err.__traceback__)
        # limit frame to command module
        try: filename = stack[1][0]
        except IndexError:
            filename = "?"
        line_number = "?"
        for frame in stack:
            if frame[0] == filename:
                line_number = frame[1]
            elif line_number != "?":
                break
        where = "(%s:%s)" % (os.path.basename(filename), line_number)
        result = python + " " + where

    if result is not None:
        if not isinstance(result, str):
            print("Error: expected str, got %s" % type(result))
            return
        result = result.encode("utf-8", "replace")
        sys.stdout.buffer.write(result + b"\n")
        sys.stdout.flush()

@public
def request(*args, **kargs):
    # Save PEP 3122!
    if "." in __name__:
        from . import web
    else:
        import web
    return web.request(*args, **kargs)

@public
def setup(function):
    function.saxo_setup = True
    if not hasattr(function, "saxo_deps"):
        function.saxo_deps = []
    return function

@public
def script(argv):
    # Save PEP 3122!
    if "." in __name__:
        from .script import main
    else:
        from script import main

    main(argv, version)



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import re
import saxo

regex_link = re.compile(r"(http[s]?://[^<> \"\x01]+)[,.]?")

@saxo.event("PRIVMSG")
def link(irc):
    search = regex_link.search(irc.text)
    if search:
        if irc.sender.startswith("#"):
            irc.client("link", irc.sender, search.group(1))



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import os.path
import random
import sys

# Save PEP 3122!
if "." in __name__:
    from . import common
    from .saxo import path as saxo_path
else:
    import common
    from saxo import path as saxo_path

DEFAULT_CONFIG = """\
# See docs/config.md for more options

[server]
    host = irc.freenode.net
    port = 6667

[client]
    nick = saxo%05i
    channels = ##saxo #test
    prefix = .
"""

E_DIRECTORY_EXISTS = """
You tried to make a directory that already exists. This is usually caused by
saxo already existing through a previous creation command. It may also be
caused by a path conflict, for example if you thought that you needed to
manually create the saxo directory beforehand.

This script won't write new material into any existing directories in case of
data corruption.
"""

E_UNMAKEABLE_DIRECTORY = """
The attempt to make this directory recursively, creating any necessary parent
directories in the process, failed. Check permissions on all parent directories
to make sure that we can write here. Also make sure that the parent path is a
directory.
"""

E_UNWRITEABLE_CONFIG = """
The saxo config file could not be written. Check that it exists, that it's a
regular file, and that you have adequate permissions to write on all parent
directories.
"""

def default(base=None):
    if base is None:
        base = os.path.expanduser("~/.saxo")

    if os.path.isdir(base):
        common.error("the directory `%s` already exists" % base,
            E_DIRECTORY_EXISTS)

    try: os.makedirs(base)
    except Exception as err:
        common.error("could not create the directory `%s`" % base,
            E_UNMAKEABLE_DIRECTORY, err)

    config = os.path.join(base, "config")
    try:
        with open(config, "w", encoding="ascii") as f:
            # Generates a random bot name, from saxo00000 to saxo99999 inclusive
            f.write(DEFAULT_CONFIG % random.randrange(0, 100000))
    except Exception as err:
        common.error("could not write the config file `%s`" % config,
            E_UNWRITEABLE_CONFIG, err)

    plugins = os.path.join(base, "plugins")
    os.mkdir(plugins)

    commands = os.path.join(base, "commands")
    os.mkdir(commands)

    common.populate(saxo_path, base)

    print("Created %s" % config)
    print("Modify this file with your own settings, and then run:")
    print("")
    if base:
        print("    %s start %s" % (sys.argv[0], base))
    else:
        print("    %s start" % sys.argv[0])

    return True



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

# python3 setup.py sdist --formats=bztar
# python3 setup.py bdist_wheel

import os.path
import sys

def error(*messages):
    for message in messages:
        print(message, file=sys.stderr)
    sys.exit(1)

try: from setuptools import setup
except ImportError as err:
    if "bdist_wheel" in sys.argv:
        error("Option bdist_wheel requires setuptools, which can't be found")
    from distutils.core import setup

try: import sqlite3
except ImportError:
    error("Error: sqlite3 is not installed",
          "Please build Python against the sqlite libraries")
else:
    if not sqlite3.threadsafety:
        error("Error: Your sqlite3 is not thread-safe")

# Do wheels invoke setup.py?
if sys.version_info < (3, 3):
    error("Error: Requires Python 3.3 or later")

def update_version():
    if os.path.isfile("version"):
        with open("version", "r", encoding="ascii") as f:
            version = f.read().rstrip("\r\n")

        if os.path.isdir(".git") and ("sdist" in sys.argv):
            major, minor, serial = [int(n) for n in version.split(".")]
            version = "%s.%s.%s" % (major, minor, serial + 1)

            with open("version", "w", encoding="ascii") as f:
                f.write(version)
    else:
        error("Error: No saxo version file found")

    return version

# http://stackoverflow.com/questions/4384796
# http://packages.python.org/distribute/

if __name__ == "__main__":
    README = "https://github.com/sbp/saxo/blob/master/README.md"

    setup(
        name="saxo",
        version=update_version(),
        author="Sean B. Palmer",
        url="http://inamidst.com/saxo/",
        description="Quick and flexible irc bot, extensible in any language",
        long_description="Documented in `@sbp/saxo/README.md <%s>`_" % README,
        license="Apache License 2.0",
        packages=["saxo"],
        package_dir={"saxo": "."},
        package_data={"saxo": [
            "README.md", "NOTICE", "version",
            "commands/*", "plugins/*", "test/*"
        ]},
        scripts=["saxo"],
        platforms="Linux and OS X",
        classifiers=[
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX",
            "Programming Language :: Python :: 3"
        ]
    )



# http://inamidst.com/saxo/
# Created by Sean B. Palmer

import saxo

@saxo.event("PRIVMSG")
def exclamation(irc):
    if irc.text == irc.config["nick"] + "!":
        irc.say(irc.nick + "!")

# TODO: Move this into prefix.py?
@saxo.event("PRIVMSG")
def prefix(irc):
    if irc.text == irc.config["nick"] + ": prefix?":
        irc.reply('"' + irc.config["prefix"] + '"')
